package experiment.event.topology;

import experiment.ExperimentState;
import experiment.entity.Link;
import experiment.entity.Node;
import experiment.event.Event;
import experiment.event.capability.Directive;
import walker.exceptions.badSyntaxException;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 11, 2010
 */
public class LinkEvent extends Event{

    private boolean status;
    private String sourceNodeId;
    private String targetNodeId;

    /**
     * Default constructor for Link Topology events.
     *
     * @param timestamp the Event timestamp
     * @param sid the link's source id
     * @param tid the link's target id
     * @param status true to enable the entity, false to disable it
     */
    public LinkEvent(long timestamp, String sid, String tid, boolean status) {
        this.timestamp = timestamp;
        this.sourceNodeId = sid;
        this.targetNodeId = tid;
        this.status = status;
    }


   /**
     * Applies this event to the experiment.
     *
     * @param state the experiment state where this event belong to.
     * @return true if event was successfully applied.
     */
    public boolean apply(ExperimentState state) throws badSyntaxException {

       // check that this is a valid linkEvent.
        if (sourceNodeId == null || targetNodeId == null){
            return false;
        }

        // Try to index the Node.
        Node sourceNode = state.indexNode(sourceNodeId);
        Node targetNode = state.indexNode(targetNodeId);

        /*
         * If the an Entity doesn't exist, add it to the experiment setup.
         */
        if (sourceNode == null){
            sourceNode = new Node(state, sourceNodeId, null, null, null, null, null);
            state.add(sourceNode);
        }

        if (targetNode == null) {
            targetNode = new Node(state, targetNodeId, null, null, null, null, null);
            state.add(targetNode);
        }

        // Try to index the Link.
        Link link = sourceNode.indexLink(targetNodeId);
        if (link == null){
            link = new Link(state, sourceNode, targetNode, null, null, null);
        }

       // If there has been a change
       if (status!=link.isEnabled()){
           // update the link's status.
           link.setStatus(status);
           // update the Link3d status
           link.getEntity3d().updateStatus();
       }

       // Update the state's time
       state.setTime(timestamp);
       return true;

   }

    /**
     * Dummy method.
     *
     * @param data the data to be added.
     */
    public void addData(Directive data) {
    }
}