package experiment.property;

import experiment.Setup.Datatype;
import experiment.Setup;
import experiment.Setup.Unit;

import java.io.Serializable;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 10, 2010
 *
 * Utilizes objects that describe the wiseml capability tag. It can be used to add capabilities to the nodes
 * (examples are temperature, humidity, etc) and links (for example LQI, packet error rate, cost, etc).
 * The user needs to specify the name, datatype, unit and a value.
 */
public class Capability implements Serializable{

    private String name;
    private Datatype dtype;
    private Unit unit;
    private String value;

    /**
     * Default Constructor.
     *
     * @param name the name of the capability
     * @param dtype the capability's data type
     * @param unit the capability's unit type
     * @param value capability's value
     */
    public Capability(String name, Datatype dtype, Unit unit, String value) {
        this.name = name;
        this.dtype = dtype;
        this.unit = unit;
        this.value = value;
    }

    /**
     * Update the capability
     * @param value contains the new value
     */
    public void update(String value) {
        this.value = value;
    }

    /**
     * Return the capability's name
     * @return a string containing the name
     */
    public String getName() {
        return name;
    }

    /**
     * Returns the capability's Datatype
     * @return a Datatype enum type
     */
    public Setup.Datatype getDatatype() {
        return dtype;
    }

    /**
     * Returns the capability's unit type
     * @return a Unit enum type
     */
    public Setup.Unit getUnit() {
        return unit;
    }

    /**
     * Returns the capability's value
     * @return s String containing the value
     */
    public String getValue() {
        return value;
    }
}
