package experiment.property;

import javax.vecmath.Vector3d;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 9, 2010
 *
 * A position in the 3 dimensional plane
 */
public class Position {

    // the values in the 3 axis
    private float x, y,z;

    private double phi, theta;

    /**
     * Default constructor
     *
     * @param x is given in degrees
     * @param y is given in degrees
     * @param z is given in radians
     * @param phi angle in degrees.
     * @param theta angle in degrees.
     */
    public Position(float x, float y, float z, double phi, double theta){
        this.x = x;
        this.y = y;
        this.z = z;
        this.phi = phi;
        this.theta = theta;
    }

    /**
     * Returns the objects x axis value
     * @return x value in meters
     */
    public float getX() {
        return x;
    }

    /**
     * Sets the x axis coordinate value
     * @param x value in meters
     */
    public void setX(float x) {
        this.x = x;
    }

    /**
     * Returns the objects y axis value
     * @return y value in meters
     */
    public float getY() {
        return y;
    }

    /**
     * Sets the y axis coordinate value
     * @param y value in meters
     */
    public void setY(float y) {
        this.y = y;
    }

    /**
     * Returns the objects z axis value
     * @return z value in degrees
     */
    public float getZ() {
        return z;
    }

    /**
     * Sets the z axis coordinate value
     * @param z value in degrees
     */
    public void setZ(float z) {
        this.z = z;
    }


    /**
     * Returns the phi angle of the relative coordinate system
     *
     * @return the phi angle
     */
    public double getPhi() {
        return phi;
    }

    /**
     * Sets the phi angle
     * @param phi tha angle in radians
     */
    public void setPhi(double phi) {
        this.phi = phi;
    }

    /**
     * Returns the theta angle of the relative coordinate system
     *
     * @return the theta angle
     */
    public double getTheta() {
        return theta;
    }

    /**
     * Sets the theta angle
     *
     * @param theta
     */
    public void setTheta(double theta) {
        this.theta = theta;
    }

    /**
     * Converts an experiment entity position to a 3d canvas position. In order to abide to java3d's orientation of
     * the axis, y and z values need to be swapped.
     *
     * @return a vactor of three doubles, containing the rearranged axis values.
     */
    public Vector3d getVector3d(){
        return new Vector3d(x, z, y);
    }

    /**
     * Compares this position to a transform position. It is used to facilitate indexing of a Node3d when a user
     * picks it.
     *
     * @param position a three dimensional Vector containing the x, y, z double values.
     * @return true is same.
     */
    public boolean same(Vector3d position) {
        return ( near(x, position.x) && near(y, position.y) && near(z, position.z) );
    }

    /**
     * Loosly compares two values.
     *
     * @param a the first value.
     * @param b the second value.
     * @return true if the values are very close or the same.
     */
    public boolean near(double a, double b){
        return (Math.abs(a-b) <= Math.max(a,b)/10);
    }
}
