/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 25, 2010
 *
 * Orchestrates the walking of the wiseml file and the creation/update of the 3d canvas.
 */
package visual;

import experiment.ExperimentState;
// unused? import experiment.Setup;
import experiment.entity.Entity;
import experiment.entity.Node;
import experiment.entity.Link;
import walker.StaxParser;
import walker.ExtendedXMLStreamReader;
import walker.Weyes;
import walker.Persistor;
import walker.exceptions.badSetupException;
import walker.exceptions.badSyntaxException;
import walker.exceptions.endDocumentException;
import static walker.utilities.Output.print;
import walker.utilities.FileUtilities;

import javax.xml.stream.XMLStreamException;
import java.util.Date;
import java.io.IOException;

import visual.entity.Entity3d;

public class Crawler extends Thread{

    StaxParser parser = null;
    private static ExperimentState state;
    private static String fileName;
    private static ExtendedXMLStreamReader reader;
    private static long fileSize = -1;
    public Engine3d engine3d;

    // Visualizator running at fps frames per second.
    private final double fps = 25;
    // The per-frame Visualization Speed.
    private volatile double speed = 1;

    // For synchronization purposes,
    private static final Object lock = new Object();

    /**
     * The thread's main function.
     */
    public void run() {

        Crawler.fileName = Weyes.fileName;
        fileSize = FileUtilities.fileSize(fileName);

        // initialize the main state.
        state = new ExperimentState();

        try {
            parser = new StaxParser(fileName, state, -1);
            reader = parser.getReader();
        } catch (IOException e) {
            e.printStackTrace();
        }

        long timer;

        // Create the 3d generated environment.
        engine3d = new Engine3d();
        Weyes.fe.setTime(0);
        
        print("#### Visualizing experiment #### \nspeed: " + speed + ", fps: " + fps + "\n");

        // Loop until the user exits.
        while (!engine3d.exit){

            // Loop while player allows it.
            while (engine3d.player.run){

                // time this
                timer = new Date().getTime();

                try {
                    // Walk the wiseml file while updating the graph.
                    parser.walk(state.getXMLReader(), state, speed, fps);
                } catch (XMLStreamException e) {
                  e.printStackTrace();
                    // Create a file with the syntax errors that the walker found.
                    Persistor.persistErrors();
                } catch (badSetupException e) {
                    e.printStackTrace();
                } catch (badSyntaxException e) {
                    e.printStackTrace();
                    // Create a file with the syntax errors that the walker found.
                    Persistor.persistErrors();
                } catch (endDocumentException e) {
                    print("end of the file reached!\nuse STOP || SEEK to continue.");
                    engine3d.player.run = false;
                }

                updateGui();
                

                // Match the parsing time with the visualization time by sleeping.
                long sleepTime = (long)(1000/fps) - (new Date().getTime()-timer);
//                print(sleepTime);
                if (sleepTime > 20){
                    try {
                        Thread.sleep(sleepTime);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }

            }

            // Handle stop and seek.
            if (engine3d.player.reset){

                // Find out which entity is currently monitored
                Entity monitoredEntity = Weyes.fe.entity;

                // Either seek,
                if (engine3d.player.seeked > 0){
                    print("\nSeeking inputStream..");
                    seek(engine3d.player.seeked);
                    engine3d.player.seeked = -1;
                    updateFrontEnd(monitoredEntity);

                // Or restart.
                } else if (engine3d.player.seeked == 0){
                    print("\nReseting inputStream..");
                    reset();
                    // Update the Sliders Indication.
                    engine3d.player.setSlider(0);
                    updateFrontEnd(monitoredEntity);
                    Weyes.fe.setTime(0);
                }

                engine3d.player.reset = false;
                // Create a file with the syntax errors that the walker found.
                Persistor.persistErrors();
            }

            // Await an update.
            synchronized (lock){
                try {
                    lock.wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

        }
    }

    /**
     * Change the visualization Speed. The speed is translated to the per frame speed by division with fps.
     *
     * @param speed the new speed value.
     */
    public void setSpeed(long speed) {
        if (speed>0 && speed < 8000000){
            this.speed = speed/fps;
        }
    }

    /**
     * After a reset or a seek operation, the FrontEnd needs to be updated. The new monitored
     * entity will be the same entity with the old one but will correspond to a different Entity instance.
     *
     * @param oldMonitoredEntity the Entity that FrontEnd monitored before the seek/reset operation.
     */
    private void updateFrontEnd(Entity oldMonitoredEntity) {

        // Match the previous monitored entity to the current one.
        if (oldMonitoredEntity != null){
            Entity newMonitoredEntity;
            if (oldMonitoredEntity.isNode()){
                Node n = (Node)oldMonitoredEntity;
                newMonitoredEntity = state.indexNode(n.getId());
            } else {
                Link l = (Link)oldMonitoredEntity;
                newMonitoredEntity = state.indexNode(l.getSource().getId()).indexLink(l.getTarget().getId());
            }

            // Change the Entity instance that FrontEnd monitores.
            Weyes.fe.changeViewedEntity(newMonitoredEntity);
        }
    }

    /**
     * Updates the GUI to show the visualization progress.
     */
    private void updateGui() {

        double time = state.getTime();
        // Update the Player.
        engine3d.player.setSlider(time);
//        engine3d.player.incrementProgress(time);
        // Update the FrontEnd.
        Weyes.fe.setTime(time);
    }

    /**
     * Tests animation by creating random events.
     */
    private void test(){
        Tester t = new Tester(state, engine3d.player);
        t.start();
    }

    /**
     * Creates a new xmlParser and reparses the wiseml from the beginning.
     * This method is called when a user stops the visualization.
     */
    public void reset(){
        try {
            // initialize the main state.
            state = new ExperimentState();
            try {
                reader.close();
            } catch (XMLStreamException e) {
                e.printStackTrace();
            }
            parser = new StaxParser(fileName, state, -1);
            reader = parser.getReader();
             // reassign entity3ds to the seeked state's Nodes and Links
            Entity3d.rewire(state);
            print("Visualization Reset");
        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    /**
     * Creates a new xmlParser and parses the wiseml file from a certain position in the file.
     * It is called when a users seeks by moving the slider.
     *
     * @param seeked the starting timestamp.
     */
    public void seek(double seeked){
        try {
            double closerTimestamp = Persistor.getCloserTimestamp(seeked);
            ExperimentState retrievedState = Persistor.retrieve(closerTimestamp);
            if (retrievedState != null) {
                state = retrievedState;
                parser = new StaxParser(fileName, state, seeked);
                reader = parser.getReader();
                print("Visualization Seeked!");
                
            } else {
                print("No snapshot exists that satisfies seeking conditions.");
                print("Take Snapshots if you want to seek.");
//                print("Visualizing from the beggining of the file.");
//                reset();
//                engine3d.player.setSlider(0);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * @return the wiseml file path in a String format
     */
    public static String getFileName() {
        return fileName;
    }

    /**
     * Every time that the position is changed by the user, either with by stopping or seeking the visualization
     * process, a new xml stream reader is constructed. This method returns that xml stream reader.
     *
     * @return the XML Stream reader currently associated with the crawler.
     */
    public static ExtendedXMLStreamReader getReader() {
        return reader;
    }

    /**
     * The 'state' object contains the experiment state for the current visualization. It is updated in every parsed
     *  event.
     * @return the experiment State that contains all the data for the current visualization.
     */
    public static ExperimentState getExperimentState(){
        return state;
    }

    /**
     * @return the size of the uncompressed wiseml file.
     */
    public static long getFileSize() {
        return fileSize;
    }

    /**
     * @return the lock associated with the Crawler thread.
     */
    public static Object getLock() {
        return lock;
    }

    /**
     *
     * @return the Per-Second Visualization Speed.
     */
    public long getSpeed() {
        return (long)(speed*fps);
    }
}
