package visual;

import com.sun.j3d.utils.behaviors.keyboard.KeyNavigatorBehavior;
import com.sun.j3d.utils.behaviors.mouse.MouseBehavior;
import com.sun.j3d.utils.behaviors.mouse.MouseRotate;
import com.sun.j3d.utils.behaviors.mouse.MouseTranslate;
import com.sun.j3d.utils.behaviors.mouse.MouseWheelZoom;
import com.sun.j3d.utils.geometry.Text2D;
import com.sun.j3d.utils.picking.PickCanvas;
import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.utils.universe.SimpleUniverse;
import experiment.ExperimentState;
import experiment.entity.Entity;
import experiment.entity.Link;
import experiment.entity.Node;
import static visual.Palette.darkgrey;
import static visual.Palette.whitegray;
import visual.entity.Entity3d;
import visual.entity.Link3d;
import visual.entity.Node3d;
import walker.Weyes;
import walker.utilities.Rendering;

import javax.media.j3d.*;
import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3f;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Collection;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Apr 1, 2010
 *
 * The Engine3d class uses java3d to draw a 3d universe and holds all things in it.
 * It also contains listeners for interaction with the user.
 */
public class Engine3d extends Frame implements MouseListener {

    private PickCanvas pickCanvas;
    public Player player;
    public volatile boolean exit = false;
    private TransformGroup RG;
    private SimpleUniverse simpleU;

    /**
     * Default Constructor. Creates the 3d Universe and occupies it with experiment Nodes and Links.
     */
    public Engine3d() {

        super("Weyes");

        // Retrieve the Experiment State that is bound to this visualization.
        ExperimentState state = Crawler.getExperimentState();

        setSize(new Dimension(Rendering.getScreenHeight(),3*Rendering.getScreenHeight()/4));
        setLocation(Rendering.getScreenHeight()/5,Rendering.getScreenHeight()/5);
        setLayout(new BorderLayout());

        GraphicsConfiguration config = SimpleUniverse.getPreferredConfiguration();

        add(Weyes.fe.getContainer(), BorderLayout.EAST);

        // VisML the Player.
        player = new Player();
        add(player.getContainer(),BorderLayout.SOUTH);

        // Create a 3d canvas and add it to the frame
        Canvas3D canvas3D = new Canvas3D(config);
        canvas3D.setMaximumSize(new Dimension(Rendering.getScreenWidth()*5/8,Rendering.getScreenHeight()));
        BorderLayout grid = new BorderLayout();
        JPanel panel = new JPanel(grid, true);
        panel.add(canvas3D);
        panel.setBorder(new EmptyBorder(15,15,5,5));
        panel.setBackground(new Color(251,251,251));
        add(panel,BorderLayout.CENTER);

        setVisible(true);

        // SimpleUniverse is a Convenience Utility class
        simpleU = new SimpleUniverse(canvas3D);
        // XXX increase the back clip distance so that we can see more nodes
        simpleU.getViewer().getView().setBackClipDistance(25.0);

        // Create the scene.
        BranchGroup scene = createSceneGraph(simpleU, state.getNodeSet());
        simpleU.addBranchGraph(scene);

        //We want to be sure we properly dispose of resources
        //this frame is using when the window is closed.  We use
        //an anonymous inner class adapter for this.
        addWindowListener(new WindowAdapter(){
            public void windowClosing(WindowEvent e) {
                dispose();
                exit = true;
                System.exit(0);
            }
        }
        );

        // Allows picking of the 3d shapes.
        pickCanvas = new PickCanvas(canvas3D, scene);
        pickCanvas.setMode(PickCanvas.GEOMETRY);
        pickCanvas.setTolerance(6);

        canvas3D.addMouseListener(this);

    }

    /**
     * Creates all things in the universe.
     *
     * @param universe is the Simple Universe where to place everything
     * @param nodes is the collection of all the experiment's Nodes.
     * @return the scene's BranchGraph.
     */
    private BranchGroup createSceneGraph(SimpleUniverse universe, Collection<Node> nodes) {

        /* Create the root of the branch graph */
        BranchGroup objRoot = new BranchGroup();

        BoundingSphere bounds = new BoundingSphere(new Point3d(), 2000.0);

        /* Create all lights */
        addLights(objRoot);

        /* Create a background */
        Background bg = new Background();
        bg.setColor(whitegray);
        bg.setApplicationBounds(bounds);
        objRoot.addChild(bg);

        /* Create the root TransformGroup */
        Transform3D rotate = new Transform3D();
        rotate.rotX(Math.PI/14.0d);
        Transform3D tempRotate = new Transform3D();
        tempRotate.rotY(Math.PI/110.0d);
        rotate.mul(tempRotate);
        Transform3D tempRotate2 = new Transform3D();
        tempRotate2.rotZ(Math.PI/270.0d);
        rotate.mul(tempRotate2);
        RG = new TransformGroup(rotate);
        RG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        RG.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        RG.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
        RG.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
        objRoot.addChild(RG);

        /* Add the axis */
        RG.addChild(new Axis3d());
        Text2D text = new Text2D("x",darkgrey, "Helvetica", 12, Font.PLAIN);
        Transform3D positionText = new Transform3D();
        positionText.set(new Vector3f(1.03f, -0.03f, -0.03f));
        TransformGroup TG0 = new TransformGroup(positionText);
        RG.addChild(TG0);
        TG0.addChild(text);

        /* Add the plane */
        RG.addChild(new Plane3d());

        /* Set up the interactive mouse rotation */
        //MouseRotate mouseRotate = new MouseRotate(MouseBehavior.INVERT_INPUT);
        MouseRotate mouseRotate = new MouseRotate();
        mouseRotate.setFactor(0.005);
        mouseRotate.setTransformGroup(RG);
        mouseRotate.setSchedulingBounds(bounds);
        RG.addChild(mouseRotate);

        /* Set up the interactive mouse translation */
        MouseTranslate mouseTranslate = new MouseTranslate();
        mouseTranslate.setTransformGroup(RG);
        mouseTranslate.setSchedulingBounds(bounds);
        RG.addChild(mouseTranslate);

        /* Set up the interactive mouse wheel zoom */
        MouseWheelZoom wheelZoom = new MouseWheelZoom();
        //MouseWheelZoom wheelZoom = new MouseWheelZoom(MouseBehavior.INVERT_INPUT);
        wheelZoom.setFactor(-1);
        wheelZoom.setTransformGroup(RG);
        //wheelZoom.setSchedulingBounds(new BoundingSphere());
        wheelZoom.setSchedulingBounds(bounds);
        RG.addChild(wheelZoom);

        /* Set up key navigation */
        TransformGroup vpTrans = universe.getViewingPlatform().getViewPlatformTransform();
        Transform3D T3D = new Transform3D();
        // define the starting camera position
        T3D.setTranslation(new Vector3f( 7f, 0f, 20.0f));
        vpTrans.setTransform(T3D);
        KeyNavigatorBehavior keyNavBeh = new KeyNavigatorBehavior(vpTrans);
        keyNavBeh.setSchedulingBounds(bounds);
        objRoot.addChild(keyNavBeh);


        /* Setting up objects */
        Entity3d.setFactor(Crawler.getExperimentState());
        
        for (Node node : nodes) {
            RG.addChild( new Node3d(node).getTG());
            Collection<Link> links = node.getLinkSet();
            for (Link link : links){
                RG.addChild(new Link3d(link).getTG());
            }
        }

        objRoot.compile();
        
        return objRoot;
    }

    /**
     * Add a new Entity during visualization. If the corresponding Entity3d exists but is hidden, it correlates the
     * two instances and shows the Entity3d. In any other case the Entity3d is created and added to the TransformGroup.
     *
     * @param entity the entity to be added.
     */
    public void addEntity(Entity entity){
        Entity3d entity3d = Entity3d.get(entity);
        if (entity3d != null){
            entity3d.show();
            entity.entity3d = entity3d;
        } else {
            BranchGroup bg = new BranchGroup();
            if (entity.isNode()){
                bg.addChild(new Node3d((Node) entity).getTG());
                Collection<Link> links = ((Node) entity).getLinkSet();
                if (links != null){
                    for (Link link : links){
                        bg.addChild(new Link3d(link).getTG());
                    }
                }
            } else {
                bg.addChild(new Link3d((Link) entity).getTG());
            }
            RG.addChild(bg);
        }
    }

    /**
     * Lights up the universe!
     *
     * @param b the scene's root element.
     */
    protected void addLights(BranchGroup b) {

        // Create bounds for the lights.
        BoundingSphere bounds = new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 100.0);

        // Create a directional light.
        Vector3f lightDir1 = new Vector3f(-20.0f, -20.0f, 0.0f);
        DirectionalLight light1 = new DirectionalLight(Palette.whitegray, lightDir1);
        light1.setInfluencingBounds(bounds);
        b.addChild(light1);

        // Create an ambient light.
        AmbientLight ambientLight1 = new AmbientLight(Palette.grey);
        ambientLight1.setInfluencingBounds(bounds);
        b.addChild(ambientLight1);

    }

    /**
     * Reset the camera position.
     * Handy if the user messes up the camera position.
     * @param mode camera mode
     */
    public void resetCamera(int mode) {
        if (mode == 1){
            TransformGroup vpTrans = simpleU.getViewingPlatform().getViewPlatformTransform();
            Transform3D T3D = new Transform3D();
            // define the starting camera position
            T3D.setTranslation(new Vector3f( 0f, 0f, 35.0f));
            vpTrans.setTransform(T3D);

            Transform3D rotate = new Transform3D();
            rotate.rotX(Math.PI/2d);
            RG.setTransform(rotate);
        } else if (mode == 2){
            TransformGroup vpTrans = simpleU.getViewingPlatform().getViewPlatformTransform();
            Transform3D T3D = new Transform3D();
            // define the starting camera position
            T3D.setTranslation(new Vector3f( 7f, 0f, 20.0f));
            vpTrans.setTransform(T3D);

            Transform3D rotate = new Transform3D();
            rotate.rotX(Math.PI/14.0d);
            Transform3D tempRotate = new Transform3D();
            tempRotate.rotY(Math.PI/110.0d);
            rotate.mul(tempRotate);
            Transform3D tempRotate2 = new Transform3D();
            tempRotate2.rotZ(Math.PI/270.0d);
            RG.setTransform(rotate);
        }

    }

    /**
     * Listens to mouse click events and uses PickCanvas to figure out where did the user click.
     *
     *  @param e is the mouse event.
     */
    public void mouseClicked(MouseEvent e) {

        pickCanvas.setShapeLocation(e);
        PickResult result = pickCanvas.pickClosest();

        if (result == null) {
            return;
        }

        javax.media.j3d.Node leaf = result.getObject();

        if (leaf == null){
            return;
        }

        // If the user clicked on a link,
        if (leaf.getName() != null && leaf.getName().equals("link")){

            String data = leaf.getUserData().toString();
            String[] nodes = data.split("-");
            Link link = Crawler.getExperimentState().indexNode(nodes[0]).indexLink(nodes[1]);

            if (link!=null){
                Weyes.fe.changeViewedEntity(link);
            }
        }

        // If the user clicked on a node,
        else if (leaf.getParent()!=null && leaf.getParent().getName()!=null && leaf.getParent().getName().equals("node")){

            // Try to retrieve the associated experiment Node
            Node node = Crawler.getExperimentState().indexNode(leaf.getParent().getUserData().toString());

            if (node != null){
                // Changed the viewed entity.
                Weyes.fe.changeViewedEntity(node);
            }

        }

    }

    /**
     * Empty mouse listener.
     *
     * @param mouseEvent a not handled event.
     */
    public void mousePressed(MouseEvent mouseEvent) {

    }

    /**
     * Empty mouse listener.
     *
     * @param mouseEvent a not handled event.
     */
    public void mouseReleased(MouseEvent mouseEvent) {

    }

    /**
     * Empty mouse listener.
     *
     * @param mouseEvent a not handled event.
     */
    public void mouseEntered(MouseEvent mouseEvent) {

    }

    /**
     * Empty mouse listener.
     *
     * @param mouseEvent a not handled event.
     */
    public void mouseExited(MouseEvent mouseEvent) {

    }
}

