package visual;

import static visual.Palette.darkgrey;

import javax.media.j3d.*;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Apr 6, 2010
 *
 * A Shape3d object that contains all functionality and data for creating a plane in 3d space.
 * The plane occupies the x and z axis.
 */
public class Plane3d extends Shape3D{

    /**
     * Default Constructor for a 2d plane in 3d canvas.
     */
    public  Plane3d(){
        this.setGeometry(createGeometry());
        this.setAppearance(createAppearance());
        this.setBounds(new BoundingSphere(new Point3d(0,0,0),2000));
    }

    /**
     * Define the appearance of the plane lines.
     *
     * @return an Appearance object.
     */
    private Appearance createAppearance() {

        Appearance app = new Appearance();

        LineAttributes line = new LineAttributes();
        line.setLineAntialiasingEnable(true);
        line.setLineWidth(0.3f);
         line.setLinePattern(LineAttributes.PATTERN_DASH);
        app.setLineAttributes(line);

        ColoringAttributes col = new ColoringAttributes(darkgrey, ColoringAttributes.SHADE_FLAT);
        app.setColoringAttributes(col);

        return app;
    }

    /**
     * Creates the geometry that defines a 2d plane.
     *
     * @return a Geometry object.
     */
    private Geometry createGeometry() {

        // draw 10 square patterns in the x,z plane beginning from (-25, -25), with steps of 5.
        float l = -25.0f;
        LineArray planeGeom = new LineArray(44, GeometryArray.COORDINATES | GeometryArray.COLOR_3);

        for(int c = 0; c < 44; c+=4){

            planeGeom.setCoordinate( c+0, new Point3f( -25.0f, 0.0f,  l ));
            planeGeom.setCoordinate( c+1, new Point3f(  25.0f, 0.0f,  l ));
            planeGeom.setCoordinate( c+2, new Point3f(   l   , 0.0f, -25.0f ));
            planeGeom.setCoordinate( c+3, new Point3f(   l   , 0.0f,  25.0f ));
            l += 5.0f;
        }
        return planeGeom;

    }

}
