package visual.entity;

import com.sun.j3d.utils.geometry.Sphere;
import com.sun.j3d.utils.geometry.Text2D;
import com.sun.media.jai.opimage.ClampCRIF;

import experiment.property.Position;
import experiment.entity.Node;
import experiment.entity.Link;
import static visual.Palette.*;

import javax.media.j3d.*;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import javax.vecmath.Color3f;

import visual.Palette;
import walker.Weyes;

import java.awt.*;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Apr 6, 2010
 *
 * Extends Entity3d abstract class by defining a viewable 3d Node.
 */
public class Node3d extends Entity3d {

    // The sphere's Radius.
    private static final float RADIUS = 0.2f;

    private Point3d center;
    private Material material;
    private PolygonAttributes polygon;
    private Alpha alpha;
    // unused? private ScaleInterpolator scale;

    private Sphere sphere;
    private Text2D text2d;
    private Text2D info2d;
    TransformGroup nodeGroup;
    TransformGroup textGroup;
    TransformGroup infotextGroup;

    private static final double TEXT_SCALE = 0.4;

    /**
     * Default Constructor for Node3d objects.
     *
     * @param node the corresponding experiment.Node.
     */
    public Node3d(Node node) {

        // call Entity3d constructor
        super(node);

        // Define the Positioning transfoms and variables.
        Vector3d vector = normalize(node.getPosition().getVector3d());
        center = new Point3d(vector);
        Transform3D translate = new Transform3D();
        translate.set(vector);

        TG = new TransformGroup();

        // initialize the Transform Group that will hold entity3d's shape.
        nodeGroup = new TransformGroup(translate);
        nodeGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        nodeGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        nodeGroup.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
        TG.addChild(nodeGroup);

        // Create a sphere.
        sphere = new Sphere(RADIUS, Sphere.GENERATE_NORMALS, 54, createAppearance());
        // Associate the sphere with the Node
        sphere.setUserData(node.getId());
        sphere.setName("node");
        // Allow the Sphere to be picked.
        sphere.setCapability(javax.media.j3d.Node.ENABLE_PICK_REPORTING);
        // add it to the Transform Group
        nodeGroup.addChild(sphere);

        // Display the Node's id as a 2d text.
        text2d = new Text2D(node.getId(), darkgrey, "Helvetica", 70, Font.PLAIN);
        text2d.setCapability(Text2D.ALLOW_APPEARANCE_WRITE);
        text2d.getAppearance().setCapability(Appearance.ALLOW_TEXTURE_WRITE);

        // Position the text above the sphere
        Transform3D textTranslation = new Transform3D();
        textTranslation.setTranslation((new Vector3d(center.getX(), center.getY() + 3*RADIUS/2, center.getZ())));
        textTranslation.setScale(TEXT_SCALE);

        // add it to the Transform Group
        textGroup = new TransformGroup(textTranslation);
        textGroup.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
        textGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

        textGroup.addChild(text2d);
        TG.addChild(textGroup);
        
        
        if(Weyes.NodeInfoCapability != null) {
	        info2d = new Text2D(Weyes.NodeInfoCapability, darkgrey, "Helvetica", 70, Font.PLAIN);
	        info2d.setCapability(Text2D.ALLOW_APPEARANCE_WRITE);
	        info2d.getAppearance().setCapability(Appearance.ALLOW_TEXTURE_WRITE);
	
	        // Position the text above the sphere
	        Transform3D infotextTranslation = new Transform3D();
	        infotextTranslation.setTranslation((new Vector3d(center.getX(), center.getY() + 3*RADIUS/2 + (text2d.getFontSize() * 0.5 * text2d.getRectangleScaleFactor()), center.getZ())));
	        infotextTranslation.setScale(TEXT_SCALE);
	
	        // add it to the Transform Group
	        infotextGroup = new TransformGroup(infotextTranslation);
	        infotextGroup.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
	        infotextGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
	        
	        infotextGroup.addChild(info2d);
	        TG.addChild(infotextGroup);
        }
    }

    /**
     * Move the 3d Entity to a new Position. NOT TESTED!
     *
     * @param p is the new experiment.entity.Position where the 3d Node is going to be moved.
     */
    public void move(Position p){
        Transform3D translate = new Transform3D();
        translate.set(p.getVector3d());
        TG.setTransform(translate);
        center = new Point3d(p.getVector3d());
    }

   /**
     * Either enlarges or shrinks the 3d node. It is used with picking and is called by FrontEnd when the monitored
     * entity changes.
     *
     * @param mode enlarge or shrink?
     */
    public void scale(SCALE mode){

       // set scaleFactor according to scale mode.
       double scaleFactor = 1;
       if (mode == SCALE.ENLARGE){
           scaleFactor = 2;
       } else if (mode == SCALE.SHRINK) {
           scaleFactor = 1;
       }

       Transform3D t = new Transform3D();

       // Scale the Node Tranform Group
       nodeGroup.getTransform(t);
       t.setScale(scaleFactor);
       nodeGroup.setTransform(t);

       // and the text Tranform Group
       textGroup.getTransform(t);
       t.setScale(scaleFactor*TEXT_SCALE);
       // reposition the text
       t.setTranslation((new Vector3d(center.getX(), center.getY() + 3*RADIUS*scaleFactor/2, center.getZ())));
       textGroup.setTransform(t);
       
       // and the text info Transform group
       if(Weyes.NodeInfoCapability != null) {
    	   infotextGroup.getTransform(t);
    	   t.setScale(scaleFactor*TEXT_SCALE);
    	   // reposition the text
    	   t.setTranslation((new Vector3d(center.getX(), center.getY() + 3*RADIUS*scaleFactor/2 + (text2d.getFontSize() * 0.5 * scaleFactor * text2d.getRectangleScaleFactor()), center.getZ())));
	       infotextGroup.setTransform(t);
       }

    }

    /**
     * Makes the 3d Node object flash red. DOESN'T WORK!
     */
    public void reden(){
        // create color interpolator
        alpha = new Alpha (1, Alpha.INCREASING_ENABLE + Alpha.DECREASING_ENABLE, 0, 0, 3000, 3000, 2000, 3000, 3000, 1000);
        ColorInterpolator colInt = new ColorInterpolator (alpha, material);
        colInt.setStartColor(purple);
        colInt.setEndColor(red);
        colInt.setSchedulingBounds(new BoundingSphere(center,10));
        TG.addChild(colInt);
    }

    /**
     * Makes a 3d Entity appear as enabled or disabled. If the Node is disabled, the Sphere has no fill. 
     */
    public void updateStatus() {
        Node node = (Node) entity;
        if (node.isEnabled()){
            polygon.setPolygonMode(PolygonAttributes.POLYGON_FILL);
            for (Link link: node.getLinkSet()){
                link.getEntity3d().updateStatus();
            }
        } else {
            polygon.setPolygonMode(PolygonAttributes.POLYGON_LINE);
            for (Link link: node.getLinkSet()){
                Link3d l3d = (Link3d)link.getEntity3d();
                l3d.updateStatus(false);
            }
        }
    }

    /**
     * @return the sphere's center Position in a Point3d vector.
     */
    public Point3d getCenter(){
        return center;
    }

    /**
     * If an entity has been constructed during experiment visualization, then in case of seeking or stop, when the
     * rewire() method is called, it may be that even though the entity3d is displayed, it has not yet been constracted.
     * In this case the entity3d is hidden until the entity is reconstructed.
     */
    public void hide() {
        sphere.getAppearance().getTransparencyAttributes().setTransparency(1);
        text2d.setString("");
    }

    /**
     * If during experiment visualization a new Entity is added and the corresponding Entity3d is hidden,
     * show the Entity3d.
     */
    public void show() {
        sphere.getAppearance().getTransparencyAttributes().setTransparency(0);
        text2d.setString(((Node)entity).getId());
    }

    /**
     * Define the appearance of the 3d Node.
     *
     * @return an Appearance object.
     */
    public Appearance createAppearance() {

        Appearance app = new Appearance();
        app.setCapability(Appearance.ALLOW_MATERIAL_WRITE);

        TransparencyAttributes trans = new TransparencyAttributes();
        trans.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
        app.setTransparencyAttributes(trans);

        ColoringAttributes ca = new ColoringAttributes(light_blue, ColoringAttributes.NICEST);
        app.setColoringAttributes(ca);

        LineAttributes line = new LineAttributes();
        line.setLineAntialiasingEnable(true);
        app.setLineAttributes(line);
        
///        material = new Material(purple, darkgrey, light_blue_purple, bright_purple,  100.0f);
//        material = new Material(white, yellow, yellow, yellow, 100.0f);
        material = new Material();
        material.setCapability(Material.ALLOW_COMPONENT_WRITE);
        material.setCapability(Material.ALLOW_COMPONENT_READ);
        app.setMaterial(material);
        // new Material(ambientColour, emissiveColour, diffuseColour, specularColour, shininess);

        polygon = new PolygonAttributes();
        polygon.setCapability(PolygonAttributes.ALLOW_MODE_WRITE);
        polygon.setPolygonMode(PolygonAttributes.POLYGON_FILL);
        polygon.setPolygonOffset(0.2f);
        app.setPolygonAttributes(polygon);

        return app;
    }


    /**
     * For 3d text. Currnetly not used.
     *
     * @param id the Node's String id.
     * @return a Geometry Object containing the 3d text.
     */
    private Geometry createTextGeometry(String id) {

        // use a customized FontExtrusion object to control the depth of the text
        Shape extrusionShape = new java.awt.geom.Line2D.Double(0, 0, 0, 0);
        FontExtrusion fontEx = new FontExtrusion(extrusionShape);
        Font3D font3d = new Font3D(new Font("SansSerif", Font.PLAIN, 1), fontEx);
        Text3D textGeometry = new Text3D(font3d, id);
        textGeometry.setAlignment(Text3D.ALIGN_CENTER);
        return textGeometry;
    }

    /**
     * For 3d text. Currently not used.
     *
     * @return The Appearance of the 3d text.
     */
    public Appearance createTextAppearance() {

        Appearance app = new Appearance();

        TransparencyAttributes trans = new TransparencyAttributes();
        trans.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
        app.setTransparencyAttributes(trans);

        Color3f col = new Color3f(0.1f, 0.1f, 0.1f);
        ColoringAttributes color = new ColoringAttributes(col, ColoringAttributes.NICEST);
        app.setColoringAttributes(color);

        polygon = new PolygonAttributes();
        polygon.setPolygonMode(PolygonAttributes.POLYGON_FILL);
        polygon.setPolygonOffset(0.2f);
        app.setPolygonAttributes(polygon);

        app.setLineAttributes(new LineAttributes(0.3f, LineAttributes.PATTERN_SOLID, true));

        Material material = new Material(grey, black, grey, white, 70f);
        app.setMaterial(material);

        return app;

    }

	public void changeColor(String data) {
		float v;
		try {
			v = Float.parseFloat(data);
			// XXX quick 'n' dirty hack to clamp msgs_rcvd from Pampa between 0 and 12
			if(v < 0) {
				v = 0;
			} else if (v > 12) {
				v = 12;
			}
		} catch (Exception e) {
			System.out.println("Malformed NodeDisplayCapability value");
			return;
		}
		
		Color3f color = new Color3f(v/12.0f, 0, 0);
		//
//		sphere.getAppearance().getColoringAttributes().setColor(color);
//		sphere.getAppearance().setMaterial(new Material(red, red, red, red, 100.0f));
//		sphere.getAppearance().getMaterial().setDiffuseColor(color);
//		sphere.getAppearance().getMaterial().setSpecularColor(color);
//		sphere.getAppearance().getMaterial().setAmbientColor(red);
		sphere.getAppearance().getMaterial().setEmissiveColor(color);
	}

	public void changeInfoText(String key, String value) {
		if(key != null && value != null) {
			info2d.setString(key + ": " + value);
		}
	}

}
